# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer o-- Rotator o-- Promax

@author: maedera and tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats

from .varimax import Varimax

class Promax():
    def __init__(self):
        self.__loading = np.array([[],[]])
        self.__fscore = np.array([[],[]])
        self.__contr = np.array([])
        self.__vmx = Varimax()


    def func(self,data,loading,nfactor,gmma=4,ndata=None,acceptable_err=1.0e-9):
        """
        decompose a given matrix without rotation
        
        Parameters
        ----------
        data: array
            preprocessed data (ex. mirror)
            feature x data (ex. gene x sample) array

        loading: array
            feature x factor (ex. gene x factor) array

        gmma: str
            determine power for indicating target matrix

        nfactor: int
            indicate the No. factors for rotation

        ndata: int
            indicate the No. of samples

        acceptable_err: float
            indicate the acceptable error in varimax rotation

        """
        ### rotation
        if nfactor > ndata:
            raise ValueError("!! nfactor is too large !!")
        L0,F0,C0 = self.__vmx.func(data=data,loading=loading,nfactor=nfactor,ndata=ndata,acceptable_err=acceptable_err)
        Lm = L0[:,:nfactor]
        Ls = L0[:,nfactor:]
        Fm = F0[:nfactor,:].T # transform matrix in general formation (n x p)
        Fs = F0[nfactor:,:]
        L = self.__promax(Lm,Lm.shape[0],Lm.shape[1],gmma) # rotated loading

        ### calculate score
        LtLinv = L.dot(np.linalg.inv(L.T.dot(L)))
        F = Fm.dot(L.T).dot(LtLinv)

        ### summary
        self.loading = np.c_[L,Ls]
        self.fscore = np.c_[F,Fs.T].T
        self.contr = C0/np.sum(C0)
        return self.loading,self.fscore,self.contr


    def __promax(self,V,n_vars,n_comp,gmma):
        """
        promax rotation algorithm
        
        Parameters
        ----------
        V: array
            rotated loading matrix (p x c)

        n_vars: int
            the No. of features

        n_comp: int
            the No. of components

        gmma: int
            indicate power for target matrix preparation

        """
        ua = V.copy()
        for i in range(n_vars):
            ua[i,:] /= np.linalg.norm(ua[i,:]) # normalization, not necessary for principal component method
        abs_ua = np.abs(ua)
        mxu = [np.max(abs_ua[:,i]) for i in range(n_comp)]
        u = ua.copy()
        for i in range(n_comp):
            u[:,i] /= mxu[i] # 最大値で割る理由は？
        C = np.array(np.sign(ua)) * (np.array(np.abs(u))**gmma)
        AtA = V.T.dot(V)
        AtAinv = np.linalg.inv(AtA)
        Q = AtAinv.dot(V.T).dot(C) # rotation matrix
        QtQinv = np.linalg.inv(Q.T.dot(Q))
        D = np.diag(np.diag(np.sqrt(QtQinv)))
        Lmbd = Q.dot(D)
        L = V.dot(Lmbd)
        return L


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer o-- Initializer o-- PC

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats

class PC():
    def __init__(self):
        self.__loading = np.array([[],[]]) # loading matrix (p x q)
        self.__fscore = np.array([[],[]]) # factor score matrix (n x q)
        self.__contr = np.array([]) # contribution


    def func(self,data,ndata=None):
        """
        decompose a given matrix without rotation
        
        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample) array
            probably preprocessed data

        ndata: int
            indicate the No. of data

        Returns
        ----------
        loading: loading matrix (novel axes)
        fscore: factor score matrix (principal components)
        contribution
        ncomp: the No. of principal components

        """
        X = data.T
        pca = PCA(svd_solver="full")
        pca.fit(X)
        self.__loading = pca.components_.T[:,:ndata]
        self.__contr = pca.explained_variance_ratio_[:ndata]
        centroid = np.c_[np.mean(data,axis=1)]
        self.__fscore = self.__loading.T.dot(data - centroid)
        return self.__loading,self.__fscore,self.__contr




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer o-- Initializer

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats

from .principal_component import PC

# abstract class
class Initializer():
    def __init__(self):
        self.__init = PC()
        self.__loading = np.array([[],[]]) # loading matrix (p x q)
        self.__fscore = np.array([[],[]]) # factor score matrix (n x q)
        self.__contr = np.array([]) # contribution


    def to_pc(self):
        """ switch to principal component method """
        self.__init = PC()
    

    def func(self,data,ndata):
        """ extract initial factors """
        self.loading,self.fscore,self.contr = self.__init.func(data,ndata)
        return self.loading,self.fscore,self.contr
    


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 25 16:01:40 2019

load data

@author: tadahaya
"""
import pandas as pd
import numpy as np
import xlrd
import csv
from pathlib import Path
import pickle


def to_pickle(data,url):
    """ to save as pickle """
    with open(url,"wb") as f:
        pickle.dump(data,f)


def read_pickle(url):
    """ to load pickle """
    with open(url,"rb") as f:
        temp = pickle.load(f)
    return temp


def load(url,ftype="csv",**kwargs):
    """
    load a file

    Parameters
    ----------
    ftype: str
        "csv","tsv","xlsx"

    sheet: int (ftype="xlsx" only)
        inicates sheet No. to be loaded

    whole: boolean (ftype="xlsx" only)
        returns a list of whole sheets data if True

    """
    file = File()
    if ftype=="tsv":
        file._2tsv()
    elif ftype=="xlsx":
        file._2xlsx()
    return file.load(url,**kwargs)


def load_line(url,ftype="tsv",**kwargs):
    """
    load a file by lines

    Parameters
    ----------
    ftype: str
        "csv" or "tsv"

    """  
    file = File()
    if ftype=="tsv":
        file._2tsv_line()
    elif ftype=="csv":
        file._2csv_line()
    else:
        raise ValueError("*** check file type!! ***")
    return file.load(url,**kwargs)


def load_multiple(url,ftype="csv",**kwargs):
    """
    load files

    Parameters
    ----------
    ftype: str
        "csv","tsv","xlsx"

    Returns
    ----------
    MultipleLoad obj

    """
    data = MultipleLoad()
    data.load(url,ftype,**kwargs)
    print("--- load multiple files ---")
    print("attributes: data, filename, path, path0")
    return data


###########################################################
class MultipleLoad():
    """ return MultipleLoad obj """
    def __init__(self):
        self._file = File()
        self.data = []
        self.filename = []
        self.path = []
        self.path0 = ""

    def _switch(self,ftype,read_line=False):
        if ftype=="tsv":
            if read_line:
                self._file._2tsv_line()
            else:
                self._file._2tsv()
        elif ftype=="csv":
            if read_line:
                self._file._2csv_line()
            else:
                self._file._2csv()
        elif ftype=="xlsx":
            self._file._2xlsx()
        else:
            raise ValueError("** check filetype!! **")

    def _get_path(self,url,ftype,extension=None):
        if extension is None:
            if ftype=="tsv":
                indicator = "*.txt"
            else:
                indicator = "*." + ftype
        else:
            indicator = "*." + extension
        p = Path(url)
        self.path = list(map(lambda x: x.as_posix(),list(p.glob(indicator))))    
        self.filename = [v.split("/")[-1].replace(indicator,"") for v in self.path]
    
    def load(self,url,ftype,extension=None,read_line=False,**kwargs):
        """
        load files
        return MultipleLoad obj

        Parameters
        ----------
        url: str
            a path indicating a folder of interest

        ftype: str
            indicate "csv","tsv", or "xlsx"
            
        extension: str
            indicate extension of files to be loaded

        sheet: int (ftype="xlsx" only)
            inicates sheet No. to be loaded

        whole: boolean (ftype="xlsx" only)
            returns a list of whole sheets data if True

        read_line: boolean (ftype="csv" or "tsv" only)
            read multiple files by line

        """
        self.path0 = url
        self._switch(ftype,read_line)
        self._get_path(url,ftype,extension)
        self.data = []
        ap = self.data.append
        for v in self.path:
            temp = self._file.load(v,**kwargs)
            ap(temp)


### not yet
class LoadLarge():
    """ load large tsv/csv reader line by line """
    def __init__(self,url=""):
        if len(url)==0:
            raise ValueError("input url!!")
        self.url = url
        self.data0 = pd.DataFrame()
        self.data = pd.DataFrame()
        self.sep = ""

    
    def _generator(self,sep="\t"):
        """ generator function """
        with open(self.url,newline="",encoding="utf-8_sig") as f:
            reader = csv.reader(f,delimiter=sep)
            for row in reader:
                yield row


    def check(self,sep="\t",num=100):
        """ to check data content """
        print("check data")
        ### loading
        gen = self._generator(sep=sep)
        lst = []
        ap = lst.append
        for i,v in enumerate(gen):
            if i < num:
                ap(v)
            else:
                break
        self.data0 = pd.DataFrame(lst)
        self.sep = sep
        return self.data0
 
    
    def load(self,sep="",focus=0,flag="",col_end=None):
        """ to read large data line by line """
        print("may consume large memory")
        ### loading
        print("loading...")
        if len(sep)!=0:            
            gen = self._generator(sep=sep)
        elif len(self.sep)!=0:
            gen = self._generator(sep=self.sep)
        else:
            gen = self._generator()
        lst = []
        ap = lst.append
        if col_end is None:
            if len(flag) > 0:                
                for v in gen:
                    if v[focus]==flag:
                        break
                    else:
                        ap(v)
            else:
                for v in gen:
                    ap(v)
        else:
            if len(flag) > 0:                
                for v in gen:
                    if v[focus]==flag:
                        break
                    else:
                        ap(v[:col_end + 1])
            else:
                for v in gen:
                    ap(v[:col_end + 1])
        self.data = pd.DataFrame(lst)
        del lst
        print("completed!!")
        return self.data


###########################################################
class File():
    def __init__(self):
        self.data = FileCsv()

    def _2csv(self):
        self.data = FileCsv()

    def _2csv_line(self):
        self.data = FileCsvLine()

    def _2tsv(self):
        self.data = FileTsv()

    def _2tsv_line(self):
        self.data = FileTsvLine()

    def _2xlsx(self):
        self.data = FileXlsx()

    def load(self,url,**kwargs):
        return self.data.load(url,**kwargs)


###########################################################
class FileXlsx():
    def load(self,url,sheet=None,whole=False):
        """
        read an excel file
        
        """
        dat = xlrd.open_workbook(url)
        sheets = dat.sheets()
        if sheet is None:
            res = []
            for v in sheets:
                try:
                    temp = self._sheet2df(v)
                    res.append(temp)
                except IndexError:
                    break
            if whole==False:
                res = res[0]
        elif type(sheet)==int:
            temp = sheets[sheet]
            res = self._sheet2df(temp)
        else:
            raise ValueError("** check sheet!! **")
        return res


    def _sheet2df(self,sheet):
        """ sheet to dataframe """
        flag = True
        i = 0
        vals = []
        ap = vals.append
        while flag:
            try:
                temp = sheet.row_values(i)
                ap(temp)
                i += 1
            except IndexError:
                flag = False
        col = vals[0]
        del vals[0]
        df = pd.DataFrame(vals,columns=col)
        return df.replace("",np.nan)


class FileCsv():
    def load(self,url,**kwargs):
        return pd.read_csv(url,index_col=0,sep=",",
                        engine='python',**kwargs) # engine is indicated for passing an initialization error 2020/4/12


class FileTsv():
    def load(self,url,**kwargs):
        return pd.read_csv(url,index_col=0,sep="\t",
                        engine='python',**kwargs) # engine is indicated for passing an initialization error 2020/4/12


class FileCsvLine():
    def load(self,url,encoding="utf-8"):
        """ load a data by line """
        with open(url,encoding=encoding) as f:
            reader = csv.reader(f,delimiter=",")
            lst = [v for v in reader]
        return pd.DataFrame(lst)


class FileTsvLine():
    def load(self,url,encoding="utf-8"):
        """ load a data by line """
        with open(url,encoding=encoding) as f:
            reader = csv.reader(f,delimiter="\t")
            lst = [v for v in reader]
        return pd.DataFrame(lst)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

@author: tadahaya
"""
import unittest
import pandas as pd
import numpy as np
import os
import sys
import math

from depro.depro import Depro

BASEPATH = os.path.dirname(os.path.abspath(__file__))

class SampleTest(unittest.TestCase):
    SETUP = None # initially loaded data

    # called when test class initialization
    @classmethod
    def setUpClass(cls):
        print('****** setUpClass method is called. ******')
        cls.SETUP = pd.read_csv(BASEPATH.replace("tests","tests\\large_grm.csv"),index_col=0)

    # called when test class end
    @classmethod
    def tearDownClass(cls):
        print('****** setDownClass method is called. ******')

    # called when a test method runs
    def setUp(self):
        self.smpl = Depro() # initialization must be done

    # called when a test method ends
    def tearDown(self):
        pass

    # wrapper for test if necessary
    def wrapper(self,X):
        if type(X)!=np.ndarray:
            return False
        elif X.shape[0]==0:
            return False
        else:
            judge = math.isnan(X[0,0])
            return not judge

    # main test
    def test_preprocess(self):
        # prepare test patterns
        test_patterns = [
            (True,0.05,True,False,True), # (arg1, arg2, ..., flag, expected result)
            (False,0.05,True,False,True), # (arg1, arg2, ..., flag, expected result)
            (True,0.05,False,False,True), # (arg1, arg2, ..., flag, expected result)
            (False,0.05,False,False,True), # (arg1, arg2, ..., flag, expected result)
            (True,0.2,True,False,True) # (arg1, arg2, ..., flag, expected result)
            ]

        ### loop for sweeping all conditions
        self.smpl.set_data(SampleTest.SETUP) # take care of how to access class variables
        for tmirror,talpha,tsphere,flag,exp in test_patterns:
            with self.subTest(mirror=tmirror,alpha=talpha,sphere=tsphere):
                print(tmirror,talpha,tsphere,flag,exp)
                if flag:
                    with self.assertRaises(exp): # give the error expected
                        self.smpl.preprocess(mirror=tmirror,alpha=talpha,sphere=tsphere) # process expected to fail
                else:
                    self.smpl.preprocess(mirror=tmirror,alpha=talpha,sphere=tsphere)
                    self.assertEqual(self.wrapper(self.smpl.get_processed()["X"]),exp)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

Organizer of data preprocessing

DePro o-- PreProcessor o-- PreProcess

@author: tadahaya
"""
import pandas as pd
import numpy as np

from .sphere import Sphere
from .mirror import Mirror
from .deg import DEGIqr,DEGPval

# abstract class
class PreProcess():
    def __init__(self):
        self.__sph = Sphere()
        self.__mirr = Mirror()
        self.__outs = []

    def get_outliers(self):
        return self.__outs

    def get_deg(self):
        """ extract DEG from input data """
        raise NotImplementedError


    def preprocess(self,data,sphere=True,mirror=True,alpha=0.05):
        """
        data preprocessing
            - unit-spherization (normalization of samples with L2 norm)
            - outlier detection with Smirnov-Grubbs test
            - mirror data

        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample) array

        sphere: boolean
            whether normalization of samples with L2 norm

        alpha: float
            indicate SG test criteria

        mirror: boolean
            whether mirror data is considered in decomposition

        """
        temp = data.copy()

        ### Mirror data
        if mirror:
            temp,self.__outs = self.__mirr.func(temp,alpha=alpha)

        ### unit-spherization
        if sphere:
            temp = self.__sph.func(temp)

        return temp


# abstract class
class DEG():
    def __init__(self):
        self.__deg = DEGIqr()

    def to_iqr(self):
        self.__deg = DEGIqr()

    def to_pval(self):
        self.__deg = DEGPval()

    def func(self,data,**kwargs):
        """
        convert dataframe of response profiles into dict of tags

        Parameters
        ----------
        data: dataframe
            feature x sample matrix

        fold: float
            determine threshold of outliers

        method: str
            "std" or "iqr"

        nmin,nmax: int
            indicate the minimum and maximum No. of genes in each tag
        
        raw: boolean
            indicates whether raw data is set or not
            if True, the data will be converted before tag making

        sep: str, default "_"
            separator for sample name
            
        position: int, default 0
            indicates position of sample name such as drug

        control: str
            indicates the control column name

        """
        return self.__deg.func(data,**kwargs)



# -*- coding: utf-8 -*-
"""

Created on Sat Mar  9 19:25:25 2019

@author: tadahaya
"""
import numpy as np
from scipy import stats

class SG():
    def __init__(self):
        pass

    def do_sg(self,data,alpha=0.05):
        """
        conduct Smirnov-Grubbs test on data (dataframe)
        return a dict containing:
            name: list of remaining data and outliers
            index: indices of remaining data and outliers
            data: arrays of remaining data and outliers
    
        Parameters
        ----------
        data: dataframe
            data

        alpha: float, default 0.05
            significance level

        """
        ind = list(data.index)
        val = data.values.flatten()
        res_ind = get_index(val,alpha=alpha)
        res_data = get_data(val,alpha=alpha)
        name_r = [ind[i] for i in res_ind[0]]
        name_o = [ind[i] for i in res_ind[1]]
        res_name = (name_r,name_o)
        key = ["name","index","data"]
        return dict(zip(key,[res_name,res_ind,res_data]))

    def get_data(self,data,alpha=0.05):
        """
        conducts Smirnov-Grubbs test on data (array)
        returns arrays of remaining data and outliers
        
        Parameters
        ----------
        data: numpy array
            data

        alpha: float, default 0.05
            significance level

        """
        x, o = list(data), []
        ap = o.append
        while True:
            n = len(x)
            if n < 3:
                break
            t = stats.t.isf(q=(alpha / n) / 2, df=n - 2)
            tau = (n - 1) * t / np.sqrt(n * (n - 2) + n * t * t)
            i_min, i_max = np.argmin(x), np.argmax(x)
            myu, std = np.mean(x), np.std(x, ddof=1)
            i_far = i_max if np.abs(x[i_max] - myu) > np.abs(x[i_min] - myu) else i_min
            tau_far = np.abs((x[i_far] - myu) / std)
            if tau_far < tau: break
            ap(x.pop(i_far))
        return (np.array(x), np.array(o))


    def get_index(self,data,alpha=0.05):
        """
        conducts Smirnov-Grubbs test on data (array)
        returns indices of remaining data and outlier ones
        
        Parameters
        ----------
        data: numpy array
            data

        alpha: float, default 0.05
            significance level

        Returns
        ----------
        rem: remained
        o: outliers

        """
        x, o = list(data), []
        y = list(data)
        ap = o.append
        while True:
            n = len(x)
            if n < 3:
                break
            t = stats.t.isf(q=(alpha / n) / 2, df=n - 2)
            tau = (n - 1) * t / np.sqrt(n * (n - 2) + n * t * t)
            i_min, i_max = np.argmin(x), np.argmax(x)
            myu, std = np.mean(x), np.std(x, ddof=1)
            i_far = i_max if np.abs(x[i_max] - myu) > np.abs(x[i_min] - myu) else i_min
            tau_far = np.abs((x[i_far] - myu) / std)
            if tau_far < tau: break
            ap(np.where(y==x[i_far])[0][0])
            del x[i_far]
        rem = [v for v in list(range(len(data))) if v not in o]
        return (rem,o)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

@author: tadahaya
"""
import unittest
import pandas as pd
import os
import sys
import math

from depro.depro import Depro

BASEPATH = os.path.dirname(os.path.abspath(__file__))

class SampleTest(unittest.TestCase):
    SETUP = None # initially loaded data

    # called when test class initialization
    @classmethod
    def setUpClass(cls):
        print('****** setUpClass method is called. ******')
        cls.SETUP = pd.read_csv(BASEPATH.replace("tests","tests\\large_grm.csv"),index_col=0)

    # called when test class end
    @classmethod
    def tearDownClass(cls):
        print('****** setDownClass method is called. ******')

    # called when a test method runs
    def setUp(self):
        self.smpl = Depro() # initialization must be done
        self.smpl.set_data(SampleTest.SETUP)
        self.smpl.preprocess()

    # called when a test method ends
    def tearDown(self):
        pass

    # wrapper for test if necessary
    def wrapper(self,df):
        if type(df)!=pd.core.frame.DataFrame:
            return False
        elif df.shape[0]==0:
            return False
        else:
            head = df.head(1)
            judge = math.isnan(head.iat[0,0])
            return not judge


    # initializer
    def initialize(self):
        nfactor = self.smpl.estimate(method="parallel",perm=3,scree_plot=True)
        L,F,C = self.smpl.decompose(initialize="principal_component",rotate="varimax",nfactor=nfactor)
        df_deg = pd.read_csv(BASEPATH.replace("tests","tests\\gep.csv"),index_col=0)
        deg = self.smpl.prep_deg(df_deg)
        return deg,L


    # main test
    def test_exdecompose(self):
        deg,L = self.initialize()

        # prepare test patterns
        test_patterns = [
            (deg,None,False,True), # (arg1, arg2, ..., exception flag, expected result)
            (deg,L,False,True), # (arg1, arg2, ..., exception flag, expected result)
            (dict(),L,True,ValueError) # (arg1, arg2, ..., exception flag, expected result)
            ]

        ### loop for sweeping all conditions
        for tdeg,tL,flag,exp in test_patterns:
            with self.subTest(deg=tdeg,loading=tL):
                print(tdeg,tL,flag,exp)
                if flag:
                    with self.assertRaises(exp):
                        self.smpl.decompose_ex(deg=tdeg,loading=tL)
                else:                   
                    self.assertEqual(self.wrapper(self.smpl.decompose_ex(deg=tdeg,loading=tL)),exp)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats

from .initializer.initializer import Initializer
from .rotator.rotator import Rotator

class Decomposer():
    def __init__(self):
        self.__init = Initializer()
        self.__rot = Rotator()
        self.__loading = np.array([[],[]])
        self.__fscore = np.array([[],[]])
        self.__contr = np.array([])

    ### data control
    def set_decomposed(self,loading=None,fscore=None,contribution=None):
        """ set external decomposed data """
        if loading is not None:
            self.__loading = loading
        if fscore is not None:
            self.__fscore = fscore
        if contribution is not None:
            self.__contr = contribution

    def get_decomposed(self):
        return {"loading":self.__loading,"fscore":self.__fscore,"contribution":self.__contr}

    ### initilizer control
    def to_pc(self):
        """ switch to principal component method """
        self.__init.to_pc()

    ### rotator control
    def to_varimax(self):
        """ switch to Varimax rotation """
        self.__rot.to_varimax()

    def to_promax(self):
        """ switch to Promax rotation """
        self.__rot.to_promax()

    ### decomposition
    def decompose(self,data,nfactor=None,ndata=None,rotate=True,acceptable_err=1.0e-7):
        """
        decompose a given matrix
        
        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample) array

        nfactor: int
            indicate the No. factors for rotation

        nsample: int
            indicate the No. of samples

        """
        ### initialize
        X = data.copy()
        if ndata is None:
            ndata = data.shape[1]
        L,F,C = self.__init.func(data=X,ndata=ndata)

        ### rotation
        if rotate:
            if (nfactor is None) or (nfactor==0):
                raise ValueError("!! Indicate the No. of factors (nfactor) !!")
            elif nfactor > len(C):
                raise ValueError("!! Inappropriate nfactor: nfactor should be less than the No. of components in initial factor prep!!")
            else:
                L,F,C = self.__rot.func(data=X,
                loading=L,nfactor=nfactor,ndata=ndata,acceptable_err=acceptable_err)

        self.__loading = L
        self.__fscore = F
        self.__contr = C
        return self.__loading,self.__fscore,self.__contr # shape: feature x contracted, contracted x data, contracted


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- ExDecomposer

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from scipy import stats

from enan.connect import Connect

class ExDecomposer():
    def __init__(self):
        self.__analyzer = Connect()
        self.__loading = np.array([[],[]]) # feature x contracted
        self.__deg = dict() # {term:tuples of up/down tags}
        self.__res = pd.DataFrame()
        

    def set_loading(self,data,trim=True,key="V"):
        """
        load rotated loading
        
        Parameters
        ----------
        data: dataframe
            feature x contracted dataframe derived from rotation

        trim: boolean
            whether loading is trimmed based on the indicated key

        key: str
            indicates the keyword for trimming

        """
        if trim:
            data = data.loc[:,data.columns.str.contains(key)]
        self.__loading = data


    def set_deg(self,data):
        """
        load DEG of chemicals of interest
        
        Parameters
        ----------
        data: dict
            a dictionary like {term:tuples of up/down tags}
            the terms correspond to chemicals of interest
        
        """
        self.__deg = data


    def get_res(self):
        """ result getter """
        return self.__res


    ### decomposition
    def decompose(self):
        """
        decompose a given matrix
        
        Parameters
        ----------
        species: str
            indicates the species to be analyzed

        """
        if len(self.__deg)==0:
            raise ValueError("!! set_deg() before this process !!")
        self.__analyzer.fit(self.__deg)
        temp = self.__loading
        self.__res = self.__analyzer.calc(temp).T
        return self.__res


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator o-- MAPc

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')

from tqdm import trange

# concrete method
class MAPc():
    def __init__(self):
        self.nfactor = 0


    def get_contr(self):
        raise NotImplementedError("!! MAP estimation does not calculate contribution !!")


    def get_nfactor(self,X,monotonic_check=5,thresh_pcorr=0.0001,fourth=False,**kwargs):
        """
        estimate the factor No. based on minimal average partial correlation (MAP)

        Parameters
        ----------
        X: 2D array
            data x feature matrix
            should be normalized

        monotonic_check: int
            indicate the upper limit of monotonic increase of the difference

        thresh_pcorr: float
            indicate partial correlation low enough to end calculation

        fourth: boolean
            whether fourth power is employed in difference calculation
            the fourth option returns more precise estimation in general but takes long time

        """
        ### preparation
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]
        R = X.T.dot(X)/ndata

        ### calculate loading
        U,S,V = np.linalg.svd(X/np.sqrt(ndata),full_matrices=False)
        loading = V.T.dot(np.diag(S)) # V is normalized, thus multiply singular values

        ### calculate average partial correlation
        num = 2
        if fourth:
            num = 4
        apcs = [1.0]
        ap = apcs.append
        count = 0
        for i in trange(loading.shape[1] - 1):
            temp = loading[:,:i + 1]
            pcorr = R - temp.dot(temp.T) # partial correlation
            if np.max(pcorr) > thresh_pcorr:
                d = 1/np.sqrt(np.diag(pcorr))
#                pcorr = np.diag(d).dot(pcorr.dot(np.diag(d))) # slower than __dX
                pcorr = self.__dX(d,pcorr)
                apc = (np.sum(pcorr**num) - nfeat)/(nfeat*(nfeat - 1))
                if apc > apcs[-1]: # check monotonic increase
                    count += 1
                    if count >= monotonic_check:
                        break
                else:
                    count = 0
                ap(apc)
            else:
                break
        minpc = np.min(apcs)
        minidx = apcs.index(minpc)
        return minidx
    

    def __dX(self,d,X):
        return d*X*np.c_[d]


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 10 10:39:56 2018

normalization code for data matrix

ver 200319

@author: tadahaya
"""
import pandas as pd
import numpy as np
from scipy.stats import rankdata
import time


def z_array(x):
    """
    to calculate z scores
    
    Parameters
    ----------
    x: a numpy array
        a numpy array to be analyzed
        
    """
    myu = np.mean(x,axis=0)
    sigma = np.std(x,axis=0,ddof=1)
    return (x - myu)/sigma    


def z_pop(x,axis=0,drop=True):
    """
    to calculate z scores from dataframe
    the scores employ population control
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    axis: 0 or 1
        whether z scores are calculate in column or row

    drop: boolean
        whether drop inf and nan
        
    """
    if axis==0:
        myu = np.mean(x.values,axis=0)
        sigma = np.std(x.values,axis=0,ddof=1)
    else:
        myu = np.c_[np.mean(x.values,axis=1)]
        sigma = np.c_[np.std(x.values,axis=1,ddof=1)]
    df = pd.DataFrame((x.values - myu)/sigma)    
    df.index = x.index
    df.columns = x.columns
    if drop:
        df = df.replace(np.inf,np.nan)
        df = df.replace(-np.inf,np.nan)
        df = df.dropna()
    return df


def z(x,control="",drop=True):
    """
    to calculate z scores based on control data from dataframe
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    control: string, default ""
        indicates the control column name

    drop: boolean
        whether drop inf and nan
        
    """
    if len(control) > 0:
        print("control column name: {0}".format(control))
        con = x.loc[:,x.columns.str.contains(control)]        
        n = len(con.columns)
        print("control column No.: {0}".format(n))
        if n < 3:
            print("<< CAUTION >> control columns are too few: population control was employed")
            return z_pop(x,axis=1,drop=drop)
        else:
            myu = np.c_[np.mean(con.values,axis=1)]
            sigma = np.c_[np.std(con.values,axis=1,ddof=1)]
            x = x.loc[:,~x.columns.str.contains(control)]
            df = pd.DataFrame((x.values - myu)/sigma)    
            df.index = x.index
            df.columns = x.columns
            if drop:
                df = df.replace(np.inf,np.nan)
                df = df.replace(-np.inf,np.nan)
                df = df.dropna()
            return df
    else:
        print("<< CAUTION >> no control columns: population control was employed")
        return z_pop(x,axis=1,drop=drop)


def madz_array(x):
    """
    to calculate MAD Z
    
    Parameters
    ----------
    x: a numpy array
        a numpy array to be analyzed
        
    """
    med = np.median(x,axis=0)
    mad = np.median(np.abs(x - med),axis=0)
    return (x - med)/(1.4826*mad)


def madz_pop(x,axis=0,drop=True):
    """
    to calculate MAD Z from dataframe
    the scores employ population control
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    axis: 0 or 1
        whether MAD Z are calculate in column or row

    drop: boolean
        whether drop inf and nan
        
    """
    if axis==0:
        med = np.median(x.values,axis=0)
        mad = np.median(np.abs(x.values - med),axis=0)
    else:
        med = np.c_[np.median(x.values,axis=1)]
        mad = np.c_[np.median(np.abs(x.values - med),axis=1)]
    df = pd.DataFrame((x.values - med)/(1.4826*mad))
    df.index = x.index
    df.columns = x.columns
    if drop:
        df = df.replace(np.inf,np.nan)
        df = df.replace(-np.inf,np.nan)
        df = df.dropna()
    return df


def madz(x,control="",drop=True):
    """
    to calculate MAD Z based on control data from dataframe
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    control: string, default ""
        indicates the control column name

    drop: boolean
        whether drop inf and nan
        
    """
    if len(control) > 0:
        print("control column name: {0}".format(control))
        con = x.loc[:,x.columns.str.contains(control)]        
        n = len(con.columns)
        print("control column No.: {0}".format(n))
        if n < 3:
            print("<< CAUTION >> control columns are too few: population control was employed")
            return madz_pop(x,axis=1,drop=drop)
        else:
            med = np.c_[np.median(con.values,axis=1)]
            mad = np.c_[np.median(np.abs(con.values - med),axis=1)]
            x = x.loc[:,~x.columns.str.contains(control)]
            df = pd.DataFrame((x.values - med)/(1.4826*mad))
            df.index = x.index
            df.columns = x.columns
            if drop:
                df = df.replace(np.inf,np.nan)
                df = df.replace(-np.inf,np.nan)
                df = df.dropna()
            return df
    else:
        print("<< CAUTION >> no control columns: population control was employed")
        return madz_pop(x,axis=1,drop=drop)


def robz_array(x):
    """
    to calculate robust z scores
    
    Parameters
    ----------
    x: a numpy array
        a numpy array to be analyzed
        
    """
    med = np.median(x,axis=0)
    q1,q3 = np.percentile(x,[25,75],axis=0)
    niqr = (q3-q1)*0.7413
    return (x - med)/niqr


def robz_pop(x,axis=0,drop=True):
    """
    to calculate robust z scores from dataframe
    the scores employ population control
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    axis: 0 or 1
        whether robust z scores are calculate in rows or columns

    drop: boolean
        whether drop inf and nan
        
    """
    if axis==0:
        med = np.median(x.values,axis=0)
        q1,q3 = np.percentile(x.values,[25,75],axis=0)
    else:
        med = np.c_[np.median(x.values,axis=1)]
        q1 = np.c_[np.percentile(x.values,25,axis=1)]
        q3 = np.c_[np.percentile(x.values,75,axis=1)]
    niqr = (q3-q1)*0.7413
    df = pd.DataFrame((x.values - med)/niqr)
    df.index = x.index
    df.columns = x.columns
    if drop:
        df = df.replace(np.inf,np.nan)
        df = df.replace(-np.inf,np.nan)
        df = df.dropna()
    return df


def robz(x,control="",drop=True):
    """
    to calculate robust z score based on control data from dataframe
    
    Parameters
    ----------
    x: a dataframe
        a dataframe to be analyzed
    
    control: string, default ""
        indicates the control column name

    drop: boolean
        whether drop inf and nan
        
    """
    if len(control) > 0:
        print("control column name: {0}".format(control))
        con = x.loc[:,x.columns.str.contains(control)]        
        n = len(con.columns)
        print("control column No.: {0}".format(n))
        if n < 3:
            print("<< CAUTION >> control columns are too few: population control was employed")
            return robz_pop(x,axis=1)
        else:
            con = x.loc[:,x.columns.str.contains(control)]  
            med = np.c_[np.median(con.values,axis=1)]
            q1 = np.c_[np.percentile(con.values,25,axis=1)]
            q3 = np.c_[np.percentile(con.values,75,axis=1)]
            niqr = (q3-q1)*0.7413
            x = x.loc[:,~x.columns.str.contains(control)]
            df = pd.DataFrame((x.values - med)/niqr)
            df.index = x.index
            df.columns = x.columns
            if drop:
                df = df.replace(np.inf,np.nan)
                df = df.replace(-np.inf,np.nan)
                df = df.dropna()
            return df
    else:
        print("<< CAUTION >> no control columns: population control was employed")
        return robz_pop(x,axis=1,drop=drop)


def quantile(df,method="median"):
    """
    quantile normalization of dataframe (variable x sample)
    
    Parameters
    ----------
    df: dataframe
        a dataframe subjected to QN
    
    method: str, default "median"
        determine median or mean values are employed as the template    

    """
    print("quantile normalization (QN)")
    df_c = df.copy() # deep copy
    lst_index = list(df_c.index)
    lst_col = list(df_c.columns)
    n_ind = len(lst_index)
    n_col = len(lst_col)

    ### prepare mean/median distribution
    x_sorted = np.sort(df_c.values,axis=0)[::-1]
    if method=="median":
        temp = np.median(x_sorted,axis=1)
    else:
        temp = np.mean(x_sorted,axis=1)
    temp_sorted = np.sort(temp)[::-1]

    ### prepare reference rank list
    x_rank_T = np.array([rankdata(v,method="ordinal") for v in df_c.T.values])

    ### conversion
    rank = sorted([v + 1 for v in range(n_ind)],reverse=True)
    converter = dict(list(zip(rank,temp_sorted)))
    converted = []
    converted_ap = converted.append
    print("remaining conversion count")    
    for i in range(n_col):
        transient = [converter[v] for v in list(x_rank_T[i])]
        converted_ap(transient)
        rem = n_col - i
        print("\r"+str(rem),end="")
    np_data = np.matrix(converted).T
    df2 = pd.DataFrame(np_data)
    df2.index = lst_index
    df2.columns = lst_col
    print("")
    print("use {}".format(method))
    return df2


def ts_norm(df,axis=0,ts=False):
    """
    normalization with total strength of each sample (columns)
    
    Parameters
    ----------
    df: dataframe
        a dataframe subjected to ts normalization
    
    axis: int, default 0
        determine direction of normalization, row or column
        0: normalization in column vector
        1: normalization in row vector
    
    ts: boolean, default "False"
        whether total strength array is exported or not    

    """
    if axis==0:
        norms = np.linalg.norm(df,axis=0)
        df2 = df/norms
    else:
        df = df.T
        norms = np.linalg.norm(df,axis=0)
        df2 = df/norms
        df2 = df2.T
    if ts:
        return df2,norms
    else:
        return df2


def consensus_sig(data,sep="_",position=0):
    """
    to generate consensus signature
    by linear combination with weightning Spearman correlation
    
    Parameters
    ----------
    data: a dataframe
        a dataframe to be analyzed
    
    sep: str, default "_"
        separator for sample name
        
    position: int, default 0
        indicates position of sample name such as drug    
    
    """
    print("generate consensus signature (time-consuming)")
    start = time.time()
    col = list(data.columns)
    ind = list(data.index)
    samples = list(set([v.split(sep)[position] for v in col]))
    samples.sort()
    rank = data.rank()
    df2 = pd.DataFrame()
    for i,v in enumerate(samples):
        lst_temp = [y.split(sep)[position]==v for y in col]
        temp = data.loc[:,lst_temp]
        col_temp = list(temp.columns)
        temp_rank = rank.loc[:,lst_temp]
        if len(col_temp)<=1:
            if i==0:
                df2 = data.loc[:,lst_temp]
            else:
                df2 = pd.concat([df2,data.loc[:,lst_temp]],axis=1,join="inner")            
        else:
            corr = np.corrcoef(temp_rank.T)
            corr_sum = np.sum(corr,axis=1) - 1
            corr = corr/np.c_[corr_sum]
            x = temp.values.T
            lst = []
            ap = lst.append
            for j in range(len(col_temp)):
                temp2 = x*(np.c_[corr[j]])
                temp2 = np.delete(temp2,j,axis=0)
                ap(np.sum(temp2,axis=0))
            x = pd.DataFrame(np.array(lst).T,index=ind,columns=col_temp)
            if i==0:
                df2 = x
            else:
                df2 = pd.concat([df2,x],axis=1,join="inner")
    end = time.time()
    h,mod = divmod(end - start,3600)
    m,s = divmod(mod,60)
    print("{0} hr {1} min {2} sec".format(int(h),int(m),round(s,4)))
    return df2


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator o-- Accumulation

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.utils.extmath import randomized_svd

class Accumulation():
    def __init__(self):
        self.nfactor = 0
        self.__contr = np.array([]) # for plot

    def get_contr(self):
        return self.__contr,np.array([])

    def get_nfactor(self,X,accumulation=0.8,show_nfactor=True,**kwargs):
        """
        estimate the factor No. from cumulative contribution

        Parameters
        ----------
        X: 2D array
            data x feature (ex. sample x gene) array

        accumulation: float
            indicate threshold of cumulative contribution
        
        """
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]        
        dat = PCAContr()
        dat.fit(X)
        self.__contr = dat.get_contr()
        contr = self.__contr/np.sum(self.__contr)
        acc_ctb = 0.0
        nfactor = 0
        for i in range(ndata):
            nfactor += 1
            acc_ctb += contr[i]
            if acc_ctb > accumulation:
                break
        self.nfactor = nfactor
        return self.nfactor

    
class PCAContr():
    def __init__(self):
        self.X = np.array([[],[]])
        self.n = 0
        self.p = 0
        self.lmd = np.array([])
    
    def fit(self,X,n_components=None):
        """
        X: 2d array
            sample x feature matrix (n x p)
            should be normalized in row (feature)
        
        """
        self.n = X.shape[0]
        self.p = X.shape[1]
        if n_components is None:
            n_components = np.min((self.n,self.p))
        U,S,V = randomized_svd(X,np.min(X.shape),transpose=False)
        lmd = S**2/self.n # the effect of sqrt(n) is reflected on sigma
        self.lmd = lmd[:n_components]

    def get_contr(self):
        return self.lmd


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

@author: tadahaya
"""
import unittest
import pandas as pd
import os
import sys
import math

from depro.depro import Depro

BASEPATH = os.path.dirname(os.path.abspath(__file__))

class SampleTest(unittest.TestCase):
    SETUP = None # initially loaded data

    # called when test class initialization
    @classmethod
    def setUpClass(cls):
        print('****** setUpClass method is called. ******')
        cls.SETUP = pd.read_csv(BASEPATH.replace("tests","tests\\large_grm.csv"),index_col=0)

    # called when test class end
    @classmethod
    def tearDownClass(cls):
        print('****** setDownClass method is called. ******')

    # called when a test method runs
    def setUp(self):
        self.smpl = Depro() # initialization must be done
        self.smpl.set_data(SampleTest.SETUP)
        self.smpl.preprocess()
        nfactor = self.smpl.estimate(method="parallel",perm=3,scree_plot=True)

    # called when a test method ends
    def tearDown(self):
        pass

    # wrapper for test if necessary
    def wrapper1(self,df):
        if type(df)!=pd.core.frame.DataFrame:
            return False
        elif df.shape[0]==0:
            return False
        else:
            head = df.head(1)
            judge = math.isnan(head.iat[0,0])
            return not judge

    def wrapper2(self,tpl):
        cL = self.wrapper1(tpl[0])
        cF = self.wrapper1(tpl[1])
        return all([cL,cF])


    # main test
    def test_decompose(self):
        # prepare test patterns
        test_patterns = [
            # (initialize,rotate,nfactor,exflag,expected)
            ("principal_component","varimax",None,False,True), # (arg1, arg2, ..., exception flag, expected result)
            ("principal_component","promax",None,False,True), # (arg1, arg2, ..., exception flag, expected result)
            ("principal_component","varimax",5,False,True), # (arg1, arg2, ..., exception flag, expected result)
            ("hoge","varimax",None,True,KeyError), # (arg1, arg2, ..., exception flag, expected result)
            ("principal_component","hoge",None,False,True) # (arg1, arg2, ..., exception flag, expected result)
            ]

        ### loop for sweeping all conditions
        for tinit,trot,tn,flag,exp in test_patterns:
            with self.subTest(initialize=tinit,rotate=trot,nfactor=tn):
                print(tinit,trot,tn,flag,exp)
                if flag:
                    with self.assertRaises(exp):
                        self.smpl.decompose(initialize=tinit,rotate=trot,nfactor=tn)
                else:                   
                    tpl = self.smpl.decompose(initialize=tinit,rotate=trot,nfactor=tn)
                    print(type(tpl[0]),type(tpl[1]),type(tpl[2]))
                    nf = self.smpl.data.nfactor
                    self.assertEqual(self.wrapper2(tpl),exp)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- PreProcessor o-- PreProcess o-- Sphere

@author: tadahaya
"""
import numpy as np

# abstract class
class Sphere():
    def __init__(self):
        pass

    def func(self,data):
        """
        unit-spherization (normalization of samples with L2 norm)

        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample) array

        """
        return data/np.linalg.norm(data,axis=0)



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Data

@author: tadahaya
"""
import pandas as pd
import numpy as np


# abstract class
class Data():
    def __init__(self):
        self.data = {"X":None,"ts":None,"index":None,"columns":None}
        self.processed = {"X":None,"outlier":None}
        self.nfactor = 0
        self.decomposed = {"loading":None,"fscore":None,"contribution":None,"ts":None}
        self.state = {"mirror":True,"sphere":True,"estimator":"parallel",
        "initializer":"principal_component","rotator":"varimax"}

    ### data loading ###
    def set_data(self,data):
        """ load data """
        if type(data)==pd.core.frame.DataFrame:
            X = data.values
            ts = np.linalg.norm(X,axis=0)
            index = list(data.index)
            columns = list(data.columns)
            self.data = {"X":X,"ts":ts,"index":index,"columns":columns}
            self.decomposed["ts"] = ts
        else:
            raise TypeError("!! data should be a dataframe !!")

    def get_data(self):
        return self.data.copy()

    def update_state(self,key,value):
        """ update state """
        if key not in list(self.state.keys()):
            raise KeyError("!! Wrong key: 'mirror', 'sphere', 'estimator', 'initializer', or 'rotator' !!")
        self.state[key] = value

    def get_state(self):
        return self.state.copy()

    ### preprocessing ###
    def set_processed(self,key,value):
        """ set preprocessed data """
        if key not in list(self.processed.keys()):
            raise KeyError("!! Wrong key: 'X' or 'outlier' !!")
        self.processed[key] = value

    def get_processed(self):
        return self.processed.copy()

    ### estimate ###
    def set_nfactor(self,n):
        """ set the No. of factors """
        self.nfactor = n

    def get_nfactor(self):
        return self.nfactor

    ### decomposition ###
    def set_decomposed(self,key,value):
        """ set loading, fscore, or contribution """
        if key not in list(self.decomposed.keys()):
            raise KeyError("!! Wrong key: 'loading', 'fscore', 'contribution', or 'ts' !!")
        self.decomposed[key] = value

    def get_decomposed(self):
        return self.decomposed

        





# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

utilities: statistics

@author: tadahaya
"""
import numpy as np

def outlier_std(mtx,fold,axis=0):
    """ calculate upper and lower values for outlier detection """
    loc = np.mean(mtx.values,axis=axis)
    scale = np.std(mtx.values,axis=axis)        
    upper = loc + fold*scale
    lower = loc - fold*scale
    return upper,lower    
    
    
def outlier_iqr(mtx,fold,axis=0):
    """ calculate upper and lower values for outlier detection """
    q3 = np.percentile(mtx.values,75,axis=axis)
    q1 = np.percentile(mtx.values,25,axis=axis) 
    scale = q3 - q1        
    upper = q3 + fold*scale
    lower = q1 - fold*scale
    return upper,lower



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer o-- Rotator o-- Varimax

@author: maedera and tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats
from math import cos,sin

class Varimax():
    def __init__(self):
        self.__loading = np.array([[],[]])
        self.__fscore = np.array([[],[]])
        self.__contr = np.array([])


    def func(self,data,loading,nfactor,ndata=None,acceptable_err=1.0e-9):
        """
        decompose a given matrix without rotation
        
        Parameters
        ----------
        data: array
            preprocessed data (ex. mirror)
            feature x data (ex. gene x sample) array

        loading: array
            feature x factor (ex. gene x factor) array

        nfactor: int
            indicate the No. factors for rotation

        ndata: int
            indicate the No. of samples

        acceptable_err: float
            indicate the acceptable error in varimax rotation

        """
        ### rotation
        if nfactor > ndata:
            raise ValueError("!! nfactor is too large !!")
        print("{} factors are subjected to Varimax rotation".format(nfactor))
        f = loading[:,:nfactor]
        rotated = self.__varimax(f,acceptable_err)
        
        ### sort factors based on variance of factor scores
        temp_score = rotated.T.dot(data)
        temp_vars = np.var(temp_score,axis=1)
        temp = dict(zip(temp_vars,rotated.T))
        s_temp = sorted(temp.items(),reverse=True)
        s_contr = np.array([v[0] for v in s_temp])
        s_rotated = np.array([v[1] for v in s_temp]).T
        self.__loading = np.c_[s_rotated,loading[:,nfactor:]]

        ### calculate factor scores
        centroid = np.c_[np.mean(data,axis=1)]
        self.__fscore = self.__loading.T.dot(data - centroid)

        ### calculate contribution
        self.__contr = s_contr/np.sum(s_contr)
        return self.__loading,self.__fscore,self.__contr # shape: feature x contracted, contracted x data, contracted


    def __varimax(self,X,acceptable_err=1.0e-7):
        """
        varimax rotation algorithm
        
        Parameters
        ----------
        X: array
            loading matrix
        
        """
        vn = X.shape[0]
        vp = X.shape[1]
        A = X.copy()
        h = [np.linalg.norm(A[i,:]) for i in range(vn)] # normalization, not necessary in principal component method
        for i in range(vn):
            A[i,:] /= h[i]
        if vp < 2:
            return A
        while True:
            ckA = A.copy()
            for i in range(vp):
                for j in range(i+1,vp):
                    x = A.T[i]
                    y = A.T[j]
                    u = x**2 - y**2
                    v = 2.0 * x * y
                    cA = np.sum(u)
                    cB = np.sum(v)
                    cC = np.sum(u**2) - np.sum(v**2)
                    cD = 2.0 * u.dot(np.c_[v])[0]
                    num = cD - 2 * cA * cB / vn
                    den = cC - (cA**2 - cB**2) / vn
                    theta4 = np.arctan2(num,den)   
                    theta = theta4/4
                    theta = theta4 / 4.0
                    tx = A.T[i] * cos(theta) + A.T[j] * sin(theta)
                    ty = -A.T[i] * sin(theta) + A.T[j] * cos(theta)
                    A.T[i] = tx
                    A.T[j] = ty
            dif = np.sum((ckA - A)**2)
            print("\r" + str(dif),end="")
            if dif < acceptable_err:
                for i in range(vn):
                    A[i,:] *= h[i]
                break
        print("")
        return A # rotated loading matrix


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

@author: tadahaya
"""
import unittest
import pandas as pd
import os
import sys
import math

from depro.depro import Depro

BASEPATH = os.path.dirname(os.path.abspath(__file__))

class SampleTest(unittest.TestCase):
    SETUP = None

    # called when test class initialization
    @classmethod
    def setUpClass(cls):
        print('****** setUpClass method is called. ******')
        cls.SETUP = pd.read_csv(BASEPATH.replace("tests","tests\\large_grm.csv"),index_col=0)

    # called when test class end
    @classmethod
    def tearDownClass(cls):
        print('****** setDownClass method is called. ******')

    # called when a test method runs
    def setUp(self):
        self.smpl = Depro() # initialization must be done
        self.smpl.set_data(SampleTest.SETUP)
        self.smpl.preprocess()

    # called when a test method ends
    def tearDown(self):
        pass

    # wrapper for test if necessary
    def wrapper1(self,df):
        if type(df)!=pd.core.frame.DataFrame:
            return False
        elif df.shape[0]==0:
            return False
        else:
            head = df.head(1)
            judge = math.isnan(head.iat[0,0])
            return not judge

    def wrapper2(self,tpl):
        cL = self.wrapper1(tpl[0])
        cF = self.wrapper1(tpl[1])
        return all([cL,cF])


    def test_estimate(self):
        # prepare test patterns
        test_patterns = [
            ("parallel",True,False,int), # (arg1, arg2, ..., exception flag, expected result)
            ("smc",True,False,int), # (arg1, arg2, ..., exception flag, expected result)
            ("map",True,False,int), # (arg1, arg2, ..., exception flag, expected result)
            ("accumulation",True,False,int), # (arg1, arg2, ..., exception flag, expected result)
            ("parallel",False,False,int), # (arg1, arg2, ..., exception flag, expected result)
            ("hoge",True,True,KeyError) # (arg1, arg2, ..., exception flag, expected result)
            ]

        ### loop for sweeping all conditions
        for tmethod,tplot,flag,exp in test_patterns:
            with self.subTest(method=tmethod,plot=tplot):
                print(tmethod,tplot,flag,exp)
                if flag:
                    with self.assertRaises(exp):
                        self.smpl.estimate(tmethod,tplot)
                else:
                    self.assertEqual(type(int(self.smpl.estimate(tmethod,tplot))),exp)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- PreProcessor o-- Mirror

@author: tadahaya
"""
import pandas as pd
import numpy as np

from .smir_grub import SG

# abstract class
class Mirror():
    def __init__(self):
        self.__sg = SG()

    def func(self,data,alpha=0.05):
        """
        convert data into mirror data
        
        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample)

        alpha: float
            the threshold for outlier detection with SG test

        """
        ts = np.linalg.norm(data,axis=0)
        remain,outs = self.__sg.get_index(ts,alpha=alpha)
        mirr = data[:,remain]
#        temp = np.c_[data,mirr*-1]
        temp = np.c_[mirr,mirr*-1]
        return temp,outs


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')

from numpy.random import normal
from scipy.stats import bernoulli
import os
import csv
import collections
from tqdm import trange
from sklearn.decomposition import PCA
from sklearn.utils.extmath import randomized_svd

from .accumulation import Accumulation
from .parallel import Parallel
from .smc import SMC
from .mapc import MAPc
from ._est_plot import ScreePlot,AccPlot

EST_METHOD = ["parallel","smc","map","accumulation"]

# organizer
class Estimator():
    """ estimate factor No. """
    def __init__(self,estimation="parallel"):
        self.__emethod = EST_METHOD
        self.__estimator = estimation
        self.__splot = ScreePlot()
        self.__aplot = AccPlot()
        if estimation=="parallel":
            self.__method = Parallel()
        elif estimation=="smc":
            self.__method = SMC()
        elif estimation=="map":
            self.__method = MAPc()
        elif estimation=="accumulation":
            self.__method = Accumulation()
        else:
            self.__method = None

    def get_emethod(self):
        return self.__emethod

    def to_parallel(self):
        self.__method = Parallel()

    def to_smc(self):
        self.__method = SMC()

    def to_map(self):
        self.__method = MAPc()

    def to_accumulation(self):
        self.__method = Accumulation()

    def get_nfactor(self,X,**kwargs):
        """
        estimate the No. of factors
        
        Parameters
        ----------
        X: 2D array
            n x p array
            should be normalized in features

        perm: int
            * for 'parallel' and 'smc'
            indicate permutation times for randomization

        monotonic_check: int
            * for 'map'
            indicate the upper limit of monotonic increase of the difference

        thresh_pcorr: float
            * for 'map'
            indicate partial correlation low enough to end calculation

        fourth: boolean
            * for 'map'
            whether fourth power is employed in difference calculation

        """
        if self.__method is None:
            raise ValueError("!! Set method for estimation !!")
        return self.__method.get_nfactor(X,**kwargs)


    def scree_plot(self,**kwargs):
        """ Scree plot """
        tcont,rcont = self.__method.get_contr()
        if len(tcont)==0:
            raise ValueError("!! run 'get_nfactor' before visualization !!")
        self.__splot.plot(tcont,rcont,**kwargs)


    def accumulation_plot(self,**kwargs):
        """ Plot accumulation """
        tcont,rcont = self.__method.get_contr()
        if len(tcont)==0:
            raise ValueError("!! run 'get_nfactor' before visualization !!")
        self.__aplot.plot(tcont,**kwargs)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

Converter

@author: tadahaya
"""

import pandas as pd
import numpy as np
from itertools import chain
from tqdm import trange
import pickle
import collections
from collections import defaultdict

class SynoDict():
    """
    Dict considering synonyms
    time consuming in dict preparation

    Parameters
    ----------
    keys: list
        representative keys for decoder
        
    synonyms: list
        a list of synonym sets [{},{},...]
        !! nan should be replaced with "" before construction !!
        
    values: list
        list of int

    """
    def __init__(self,keys=[],values=[],synonyms=[],processing=True):
        if processing:
            self.keys = keys
            lkeys = list(map(lambda x: str(x).lower(),keys))
            self.values = list(map(lambda x: int(x),values))
            self.decoder = dict(zip(self.values,lkeys))
            synonyms = [set([str(x).lower() for x in v]) for v in synonyms]
            lst = list(chain.from_iterable(synonyms))
            ol = set(k for k,v in collections.Counter(lst).items() if v > 1) | {""}
            synonyms = [v - ol for v in synonyms]
            self.synonyms = [{v} | w for v,w in zip(self.keys,synonyms)]
        else:
            self.keys = keys
            self.values = list(map(lambda x: int(x),values))
            self.decoder = dict(zip(values,keys))
            self.synonyms = synonyms
        try:
            self.sub = max(np.array(values)) + 1 # int used for substitution when KeyError occurs
        except ValueError:
            self.sub = 0
        self.__zip = list(zip(self.synonyms,self.values))
        self.not_found = dict()


    def enc(self,word):
        """ encoder """
        for v,w in self.__zip:
            if word in v:
                return w
        raise KeyError(word)


    def fix(self,obj):
        """
        return fixed dict for converting the indicate list
        
        Parameters
        ----------
        obj: list
            a list of conversion target

        substitute: str
            a word employed for indicating not found keys        

        """
        value = self.enc_list(obj)
        return dict(zip(obj,value))


    def get_not_found(self):
        """ get not found dictionary """
        return self.not_found


    def to_pickle(self,url):
        """ to save Synonym Dict """
        with open(url,"wb") as f:
            pickle.dump([self.keys,self.values,self.synonyms],f)


    def read_pickle(self,url):
        """ to load Synonym Dict """
        with open(url,"rb") as f:
            temp = pickle.load(f)
        self.keys = temp[0]
        self.values = temp[1]
        self.synonyms = temp[2]
        self.decoder = dict(zip(self.values,self.keys))
        self.__zip = list(zip(self.synonyms,self.values))
        self.not_found = dict()
        try:
            self.sub = max(np.array(temp[1])) + 1 # int used for substitution when KeyError occurs
        except ValueError:
            self.sub = 0


    def enc_list(self,target):
        """
        convert a list according to pre-defined dict

        Parameters
        ----------
        target: list
        
        """
        res = []
        ap = res.append
        nf = []
        ap2 = nf.append
        i = 0
        for v in target:
            try:
                ap(self.enc(v))
            except KeyError:
                temp = self.sub + i
                ap(temp)
                self.not_found[v] = temp
                i += 1
        self.keys += list(self.not_found.keys())
        self.synonyms += [set() for l in range(i)]
        self.values += list(self.not_found.values())
        tdic = dict(zip(list(self.not_found.values()),list(self.not_found.keys())))
        self.decoder.update(tdic)
        self.sub += i
        return res


    def dec_list(self,target):
        """ decoder for list """
        return [self.decoder[v] for v in target]


    def enc_set(self,target):
        """
        convert a set according to pre-defined dict

        Parameters
        ----------
        target: set

        substitute: str
            a word employed for indicating not found keys        
        
        """
        res = set()
        ad = res.add
        i = 0
        for v in target:
            try:
                ad(self.enc(v))
            except KeyError:
                temp = self.sub + i
                ad(temp)
                self.not_found[v] = temp
                i += 1
        self.keys += list(self.not_found.keys())
        self.synonyms += [set() for l in range(i)]
        self.values += list(self.not_found.values())
        tdic = dict(zip(list(self.not_found.values()),list(self.not_found.keys())))
        self.decoder.update(tdic)
        self.sub += i
        return res


    def dec_set(self,target):
        """ decoder for set """
        return {self.decoder[v] for v in target}


    def aggregate(self):
        """ aggregate values """
        dup = [x for x in set(self.values) if self.values.count(x) > 1]
        n = len(self.keys)
        if len(dup) > 0:
            new_keys = []
            new_syno = []
            new_idx = []
            ap = new_keys.append
            ap2 = new_syno.append
            for d in dup:
                idx = [i for i,v in enumerate(self.values) if v==d]
                syno = set(chain.from_iterable([self.synonyms[i] for i in idx]))
                n_keys = [len(self.keys[i]) for i in idx]
                ap(self.keys[np.argmin(n_keys)])
                new_idx += idx
            rem_vals = [v for v in self.values if v not in dup]
            rem_keys = [self.keys[i] for i in range(n) if i not in new_idx]
            rem_syno = [self.synonyms[i] for i in range(n) if i not in new_idx]
            self.keys = rem_keys + new_keys
            self.values = rem_vals + dup
            self.synonyms = rem_syno + new_syno


class FixedDict(SynoDict):
    """ handling conversion between names and IDs """
    def __init__(self,keys=[],values=[],synonyms=[]):
        self.keys = keys
        self.values = values
        self.decoder = dict(zip(values,keys))
        self.synonyms = synonyms
        self.sub = max(values) + 1 # int used for substitution when KeyError occurs
        self.not_found = dict()
        self.encoder = dict(zip(self.keys,self.values))


    def enc_list(self,target):
        """
        convert a list according to pre-defined dict

        Parameters
        ----------
        target: list
        
        """
        return [self.encoder[v] for v in target]


    def dec_list(self,target):
        """ decoder for list """
        return [self.decoder[v] for v in target]


    def enc_set(self,target):
        """
        convert a set according to pre-defined dict

        Parameters
        ----------
        target: set
        
        """
        return {self.encoder[v] for v in target}


    def dec_set(self,target,substitute=0):
        """ decoder for set """
        return {self.decoder[v] for v in target}


class Integrator():
    def __init__(self):
        self.ref = None
        self.encoder = dict() # integrated dict for encoding
        self.decoders = dict() # dict of dict for each decoding
        self.keys = []


    def refresh_keys(self):
        """ refresh keys """
        self.keys = []


    def make_ref(self,keys=[],values=[],synonyms=[]):
        """ prepare SynoDict """
        self.ref = SynoDict(keys,values,synonyms,processing=True)
        return self.ref


    def load_ref(self,ref):
        """
        load a reference

        Parameters
        ----------
        ref: SynoDict

        """
        self.ref = ref


    def register(self,keys=[],name=None,ref=None,drop=False):
        """
        registration of a member

        Parameters
        ----------
        name: str
            indicates the key for the data

        keys: list
            data for dictionary preparation
            !! Should be lower case !!
                    
        """
        if len(keys)==0:
            raise ValueError("!! Give keys as a list !!")
        if ref is not None:
            self.ref = ref
        if self.ref is None:
            raise ValueError("!! Load reference before this process !!")
        if name is None:
            name = "decoder" + str(len(self.keys))
        if drop:
            values = self.ref.enc_list(keys)
            nf = set(self.ref.get_not_found().keys())
            keys = [k for k in keys if k not in nf]
            values = [v for k,v in zip(keys,values) if k not in nf]
        else:
            values = self.ref.enc_list(keys)
        dic = dict(zip(keys,values))
        self.encoder.update(dic)
        dec = dict(zip(values,keys))
        self.decoders[name] = dec
        self.keys.append(name)


    def extend_encoder(self,dic:dict,name:str="EXTENDED"):
        """
        extend encoder by adding dic

        Parameters
        ----------
        name: str
            indicates the key for the data
        
        """
        self.encoder.update(dic)
        dec = dict(zip(dic.values(),dic.keys()))
        self.decoders[name] = dec
        self.keys.append(name)


    def merge_decoder(self,keys:list=[],new_name:str="MERGED"):
        """
        merge decoders with the indicated keys

        Parameters
        ----------
        keys: list
            indicates the keys to be concatenated

        new_name: str
            indicates the key for the concatenated data
        
        """
        if len(keys) < 2:
            raise ValueError("!! Few keys for concatenation !!")
        temp = self.decoders[keys[0]]
        for k in keys[1:]:
            temp.update(self.decoders[k])
        self.decoders[new_name] = temp


    def enc_list(self,target):
        """
        convert a list according to the integrated encoder

        Parameters
        ----------
        target: list
        
        """
        try:
            res = [self.encoder[v] for v in target]
        except KeyError("!! Some inputs are not found: register(drop=False) whole keys to be analyzed !!"):
            pass
        return res


    def enc_set(self,target):
        """
        convert a list according to the integrated encoder

        Parameters
        ----------
        target: set
        
        """
        try:
            res = {self.encoder[v] for v in target}
        except KeyError("!! Some inputs are not found: register(drop=False) whole keys to be analyzed !!"):
            pass
        return res


    def dec_list(self,target,key):
        """
        convert a list according to the integrated decoder

        Parameters
        ----------
        target: list
        
        """
        try:
            res = [self.decoders[key][v] for v in target]
        except KeyError("!! Some inputs are not found: register(drop=False) whole keys to be analyzed !!"):
            pass
        return res


    def dec_set(self,target,key):
        """
        convert a list according to the integrated decoder

        Parameters
        ----------
        target: set
        
        """
        try:
            res = {self.decoders[key][v] for v in target}
        except KeyError("!! Some inputs are not found: register(drop=False) whole keys to be analyzed !!"):
            pass
        return res


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator o-- SMC

@author: tadahaya
"""
import pandas as pd
import numpy as np
from numpy.random import normal
import collections
from tqdm import trange
from sklearn.utils.extmath import randomized_svd


# concrete method
class SMC():
    def __init__(self):
        self.nfactor = 0
        self.__true_contr = np.array([]) # for plot
        self.__rand_contr = np.array([]) # for plot


    def get_contr(self):
        """
        get eigen values for contribution calculation and plot
        
        Return
        ------
        true_eigen_values: eigen values derived from the actual data
        random_eigen_values: eigen values derived from the representative random data
        
        """
        return self.__true_contr,self.__rand_contr


    def get_nfactor(self,X,perm=3,**kwargs):
        """
        estimate the factor No. with diagonal SMC parallel analysis
        randomization with Normal distribution

        Parameters
        ----------
        X: 2D array
            data x feature matrix

        perm: int
            indicate permutation No. for parallel analysis
        
        """
        ### preparation
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]        
        dat = SMCContr()
        dat.fit(X)
        self.__true_contr = dat.get_contr()

        ### contribution of randomized data
        Xshape = X.shape
        myu = np.mean(X)
        uvar = np.var(X,ddof=1)
        rand_contr = []
        ap = rand_contr.append
        for i in trange(perm):
            Xrand = normal(myu,uvar,Xshape)
            
            ### get contribution
            dat = SMCContr()
            dat.fit(Xrand)
            temp_contr = dat.get_contr()
            ap(temp_contr)
        rand_contr = np.array(rand_contr)

        ### determine the threshold
        checker = self.__true_contr - rand_contr
        high = np.sum(checker > 0,axis=1)
        self.nfactor = _mode_higher(high)
        high_idx = [i for i,v in enumerate(high) if v==self.nfactor]
        self.__rand_contr = rand_contr[np.min(high_idx)]
        return self.nfactor


    def _get_nfactor_random(self,X,perm=5,**kwargs):
        """
        estimate the factor No. with parallel analysis

        Parameters
        ----------
        X: 2D array
            data x feature matrix

        perm: int
            indicate permutation No. for parallel analysis
        
        """
        ### preparation
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]        
        dat = SMCContr()
        dat.fit(X)
        self.__true_contr = dat.get_contr()

        ### contribution of randomized data
        X_c = X.copy()
        rand_contr = []
        ap = rand_contr.append
        for i in trange(perm):
            rand_data = []
            ap2 = rand_data.append
            for v in X_c:
                temp = v.copy()
                np.random.shuffle(temp)
                ap2(temp)
            Xrand = np.array(rand_data)
            
            ### get contribution
            dat = SMCContr()
            dat.fit(Xrand)
            temp_contr = dat.get_contr()
            ap(temp_contr)
        rand_contr = np.array(rand_contr)

        ### determine the threshold
        checker = self.__true_contr - rand_contr
        high = np.sum(checker > 0,axis=1)
        self.nfactor = _mode_higher(high)
        high_idx = [i for i,v in enumerate(high) if v==self.nfactor]
        self.__rand_contr = rand_contr[np.min(high_idx)]
        return self.nfactor


class SMCContr():
    def __init__(self):
        self.n = 0
        self.p = 0
        self.d2 = np.array([])
        self.smc = np.array([])
        self.lmd = np.array([])
        self.sum = 0.0

    def fit(self,X,n_components=None):
        """
        X: 2d array
            sample x feature matrix (n x p)
            should be normalized in row (feature)
        
        """
        self.n = X.shape[0]
        self.p = X.shape[1]
        if n_components is None:
            n_components = np.min((self.n,self.p))
        U,S,V = randomized_svd(self._calc_smc(X),np.min(X.shape),transpose=False) # considering natures of symmetric matrix
        del U,V
        self.lmd = S[:n_components]

    def _calc_smc(self,X):
        R = X.T.dot(X)/X.shape[0]
        Rinv = np.linalg.inv(R)
        self.d2 = 1/np.diag(Rinv)
        del Rinv
        self.smc = 1 - self.d2 # squared multiple correlation is equivalent to communalities
        return R - np.diag(self.d2)

    def get_contr(self):
        return self.lmd

def _mode_higher(array):
    """ return mode which is highest when multiple hits """
    res = collections.Counter(array).most_common()
    mod = res[0]
    res = [v[0] for v in res if v[1]==mod[1]]
    return np.max(res)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Decomposer o-- Rotator

@author: tadahaya
"""
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')
from sklearn.decomposition import PCA
from scipy import stats

from .varimax import Varimax
from .promax import Promax

# abstract class
class Rotator():
    def __init__(self):
        self.__rot = Varimax()
        self.__loading = np.array([[],[]])
        self.__fscore = np.array([[],[]])
        self.__contr = np.array([])

    def to_varimax(self):
        """ switch to varimax rotation """
        self.__rot = Varimax()

    def to_promax(self):
        """ switch to promax rotation """
        self.__rot = Promax()

    def func(self,data,loading,nfactor,ndata=None,acceptable_err=1.0e-7):
        """ rotate loading """
        L,F,C = self.__rot.func(data=data,loading=loading,nfactor=nfactor,ndata=ndata,acceptable_err=acceptable_err)
        self.__loading = L
        self.__fscore = F
        self.__contr = C
        return self.__loading,self.__fscore,self.__contr # shape: feature x contracted, contracted x data, contracted


# -*- coding: utf-8 -*-
"""
Created on Wed Dec 25 15:46:32 2019

@author: tadahaya
"""
from .depro import Depro

__copyright__    = 'Copyright (C) 2020 MIZUNO Tadahaya'
__version__      = '0.0.11'
__license__      = 'MIT'
__author__       = 'MIZUNO Tadahaya'
__author_email__ = 'tadahaya@gmail.com'


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

@author: tadahaya
"""
import unittest
import pandas as pd
import os
import sys
import math

from depro.depro import Depro

BASEPATH = os.path.dirname(os.path.abspath(__file__))

class SampleTest(unittest.TestCase):
    SETUP = None # initially loaded data

    # called when test class initialization
    @classmethod
    def setUpClass(cls):
        print('****** setUpClass method is called. ******')
        cls.SETUP = pd.read_csv(BASEPATH.replace("tests","tests\\gep.csv"),index_col=0)

    # called when test class end
    @classmethod
    def tearDownClass(cls):
        print('****** setDownClass method is called. ******')

    # called when a test method runs
    def setUp(self):
        self.smpl = Depro() # initialization must be done

    # called when a test method ends
    def tearDown(self):
        pass

    # wrapper for test if necessary
    def wrapper(self,dic):
        if type(dic)!=dict:
            return False
        elif len(list(dic.keys()))==0:
            return False
        else:
            return True

    # main test
    def test_deg(self):
        # prepare test patterns
        test_patterns = [
            (3,None,None,True,"_",0,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,10,None,True,"_",0,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,500,True,"_",0,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,None,False,"_",0,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,None,True,"-",0,"control",True,ValueError), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,None,True,"_",1,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,None,True,"_",0,"control",False,True), # (arg1, arg2, ..., exception flag, expected result)
            (3,None,None,True,"_",0,"hoge",False,True), # (arg1, arg2, ..., exception flag, expected result)
            ]

        ### loop for sweeping all conditions
        for tfold,tmin,tmax,traw,tsep,tpos,tcon,flag,exp in test_patterns:
            print(tfold,tmin,tmax,traw,tsep,tpos,tcon,flag,exp)
            with self.subTest(fold=tfold,nmin=tmin,nmax=tmax,raw=traw,sep=tsep,position=tpos,control=tcon):
                if flag:
                    with self.assertRaises(exp):
                        self.smpl.prep_deg(data=SampleTest.SETUP,fold=tfold,nmin=tmin,nmax=tmax,
                                           raw=traw,sep=tsep,position=tpos,control=tcon)
                else:
                    dic = self.smpl.prep_deg(data=SampleTest.SETUP,fold=tfold,nmin=tmin,nmax=tmax,
                                    raw=traw,sep=tsep,position=tpos,control=tcon)
                    self.assertEqual(self.wrapper(dic),exp)



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

assistant for coding

@author: tadahaya
"""

import time
import pickle

def chrono(func):
    def wrapper(*args,**kwargs):
        start = time.time()
        res = func(*args,**kwargs)
        end = time.time()
        h,rem = divmod(end - start,3600)
        m,s = divmod(rem,60)
        print("--- {0} h {1} m {2} s ---".format(int(h),int(m),s))
        return res
    return wrapper


def to_pickle(data,url):
    """ to save as pickle """
    with open(url,"wb") as f:
        pickle.dump(data,f)


def read_pickle(url):
    """ to load pickle """
    with open(url,"rb") as f:
        temp = pickle.load(f)
    return temp


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

Decomposition analysis of Profile data (DePro)

@author: tadahaya
"""
import pandas as pd
import numpy as np
from copy import deepcopy

from .data import Data
from .preprocess.preprocessor import PreProcess,DEG
from .estimator.estimator import Estimator
from .decomposer.decomposer import Decomposer
from .decomposer.ex_decomposer import ExDecomposer
from .plot.plot import RadarChart

# organizer
class Depro():
    def __init__(self):
        self.data = Data()
        self.__prep = PreProcess()
        self.__deg = DEG()
        self.__estr = Estimator()
        self.__deco = Decomposer()
        self.__xdeco = ExDecomposer()

    ### data control ###
    def set_data(self,data):
        """ set a data to be analyzed """
        self.data.set_data(data)

    def get_data(self):
        """ get data """
        return self.data.get_data()

    def get_processed(self):
        """ get processed data """
        return self.data.get_processed()

    def get_decomposed(self):
        """ get processed data """
        return self.data.get_decomposed()

    def get_state(self,show=False):
        """ get states of analysis """
        dic = self.data.get_state()
        if show:
            for k,v in dic.items():
                print("{0}: {1}".format(k,v))
        return dic

    ### preprocessing ###
    def preprocess(self,mirror=True,alpha=0.05,sphere=True,**kwargs):
        """
        data preprocessing

        Parameters
        ----------
        mirror: boolean
            indicate whether mirror data is employed
            for setting the centroid of novel space to the data origin

        alpha: float
            indicate SG test criteria

        show: boolean
            whether details are shown or not

        sphere: boolean
            whether normalization of samples with L2 norm

        """
        data = self.data.get_data()
        X = data["X"]
        Xpro = self.__prep.preprocess(X,mirror=mirror,sphere=sphere,alpha=alpha,**kwargs)
        self.data.set_processed("X",Xpro)
        if mirror:
            self.data.update_state(key="mirror",value=True)
            outs = self.__prep.get_outliers()
            self.data.set_processed("outlier",outs)
            if len(outs) > 0:
                print(">> {} samples below are excluded from the mirror".format(len(outs)))
                samples = data["columns"]
                for v in outs:
                    print("- {}".format(samples[v]))
        else:
            self.data.update_state(key="mirror",value=False)
        if sphere:
            self.data.update_state(key="sphere",value=True)
        else:
            self.data.update_state(key="sphere",value=False)


    ### estimate the factor No. ###
    def estimate(self,method="parallel",scree_plot=False,**kwargs):
        """
        data preprocessing

        Parameters
        ----------
        method: str
            indicate the estimation method for determining the No. of important factors
            "accumulation", "parallel", "smc", or "map"
            note that "smc" is a quite heavy process

        """
        X = self.data.get_data()["X"] # in estimation, z is simply calculated from raw without considering mirror  
        sphere = self.get_state()["sphere"]
        if sphere:
            X = X/self.data.get_data()["ts"]
        if method=="parallel":
            self.__estr.to_parallel()
            self.data.update_state(key="estimator",value="parallel")
            Z = X.T
        elif method=="smc":
            self.__estr.to_smc()
            self.data.update_state(key="estimator",value="smc")
            Z = _normal(X.T)
        elif method=="map":
            self.__estr.to_map()
            self.data.update_state(key="estimator",value="map")
            Z = _normal(X.T)
        elif method=="accumulation":
            self.__estr.to_accumulation()
            self.data.update_state(key="estimator",value="accumulation")
            Z = X.T
        else:
            raise KeyError("!! Wrong method: choose 'accumulation', 'parallel', 'smc', 'mapc', or 'accumulation' !!")
        nfactor = self.__estr.get_nfactor(Z,**kwargs)
        self.data.set_nfactor(nfactor)
        if scree_plot:
            try:
                self.__estr.scree_plot(**kwargs)
            except NotImplementedError:
                pass
        return nfactor


    ### decomposition ###
    def decompose(self,initialize="principal_component",rotate="varimax",nfactor=None,**kwargs):
        """
        calculation

        Parameters
        ----------
        initialize: str
            indicate the initial method for preparing factors
            "principal_component"

        rotation: str
            indicate the rotation method for reducing the features consisting axes
            "varimax" or "optimax"
            "", False, or None means no rotation (equals to conventional PCA)
        
        """
        ### initialize
        if initialize=="principal_component":
            self.__deco.to_pc()
            self.data.update_state(key="initializer",value="principal_component")
        else:
            raise KeyError("!! Wrong method: choose 'principal_component' !!")

        ### rotate
        if rotate=="varimax":
            self.__deco.to_varimax()
            self.data.update_state(key="rotator",value="varimax")
            rotate = True
        elif rotate=="promax":
            raise NotImplementedError
            # self.__deco.to_promax()
            # self.data.update_state(key="rotator",value="promax")
            # rotate = True
        else:
            print("!! No rotation: indicate 'varimax' or 'promax' when rotation is necessary !!")
            rotate = False
            self.data.update_state(key="rotator",value=None)
        X = self.data.get_processed()["X"]
        if nfactor is None:
            nfactor = self.data.get_nfactor()
        L,F,C = self.__deco.decompose(data=X,nfactor=nfactor,rotate=rotate,**kwargs)
        state = self.get_state()
        if state["mirror"]:
            d = self.data.get_data()
            X = d["X"]
            nsample = X.shape[1]
            L = L[:,:nsample]
            C = C[:nsample]
            if state["sphere"]:
                ts = d["ts"]
                X = X/ts
            F = L.T.dot(X)
        label = _prep_label(F.shape[1],len(C))
        L = pd.DataFrame(L,index=self.data.data["index"],columns=label)
        F = pd.DataFrame(F,index=label,columns=self.data.data["columns"])
        self.data.set_decomposed("loading",L)
        self.data.set_decomposed("fscore",F)
        self.data.set_decomposed("contribution",C)
        return L,F,C


    def _logical_conf(self):
        """ reconstruct data for confirmation of calculation """
        deco = self.data.get_decomposed()
        raw = self.data.get_data()
        pro = self.data.get_processed()
        L = deco["loading"].values
        F = deco["fscore"].values
        ts = deco["ts"]
        state = self.get_state()
        if state["rotator"] is None:
            centroid = np.c_[np.mean(pro["X"],axis=1)]
            recX = L.dot(F) + centroid
        elif state["rotator"]=="promax":
            raise NotImplementedError
        else:
            recX = L.dot(F)
            if not state["mirror"]:
                centroid = np.c_[np.mean(pro["X"],axis=1)]
                recX = recX + centroid
        if state["sphere"]:
            recX = recX.dot(np.diag(ts))
        del deco,L,F,ts,pro
        rec = pd.DataFrame(recX,index=raw["index"],columns=raw["columns"])
        rdata = pd.DataFrame(raw["X"],index=raw["index"],columns=raw["columns"])
        return rec,rdata


    ### decomposition of external data ###
    def prep_deg(self,data,**kwargs):
        """
        convert dataframe into dict of tags

        Parameters
        ----------
        data: dataframe
            feature x sample matrix
            In the default setting, log2 expression matrix is assumed
            If the data is already response profiles, use raw=False option

        fold: float
            determine threshold of outliers

        method: str
            "std" or "iqr"

        nmin,nmax: int
            indicate the minimum and maximum No. of genes in each tag
        
        raw: boolean
            indicates whether raw data is set or not
            if True, the data will be converted before tag making

        sep: str, default "_"
            separator for sample name
            
        position: int, default 0
            indicates position of sample name such as drug

        control: str
            indicates the control column name

        """
        idx = list(data.index)
        if self.data.decomposed["loading"] is not None:
            temp = list(self.data.decomposed["loading"].index)
            idx = sorted(list(set(idx) & set(temp)))
        elif self.data.data["index"] is not None:
            temp = self.data.data["index"]
            idx = sorted(list(set(idx) & set(temp)))
        else:
            print("!! CAUTION: DEG preparation is done without merging reference data !!")
        data = data.loc[idx,:]
        return self.__deg.func(data,**kwargs)


    def decompose_ex(self,deg=dict(),loading=None,**kwargs):
        """
        decompose the external data based on decomposed effects
        
        Parameters
        ----------
        deg: dict
            a dictionary like {term:tuples of up/down tags}
            the terms correspond to chemicals of interest

        trim: boolean
            whether loading is trimmed based on the indicated key

        key: str
            indicates the keyword for trimming

        species: str
            indicates the species to be analyzed

        """
        if len(deg.keys())==0:
            raise ValueError("!! Indicate DEG to be analyzed !!")
        if loading is None:
            self.__xdeco.set_loading(self.get_decomposed()["loading"],**kwargs)
        else:
            self.__xdeco.set_loading(loading,**kwargs)
        self.__xdeco.set_deg(deg)
        res = self.__xdeco.decompose()
        return res


    ### visualization ###
    def plot(self,score=None,focus=[],absolute=False,overlay=False,color="mediumblue",
             n_labeled=None,plot_all=False,figsize=None,savefig="",dpi=100,**kwargs):
        """
        plot scores as a radarchart
        
        Parameters
        ----------
        score: dataframe
            data to be visualized (contracted x samples)
            Both results of decompose() and decompose_ex() can be applied

        absolute: boolean
            indicates whether data is converted into absolute values before visualization

        focus: list
            a list of sample names to be visualized

        overlay: boolean
            whether plots are overlaid or not

        color: str or list
            indicates the color of plot

        n_labeled: int
            indicates the No. of labels to be plotted

        plot_all: boolean
            whether all samples are visualized or not

        limit: tuple
            determine min and max values of plot: (min,max)
        
        color: str
            determine the color of plot
        
        figsize: tupple
            determine figsize
        
        markersize: int
            determine size of markers
            
        labelsize: int
            determine size of axis-labels
            
        ticksize: int
            deterimne font size of polar axis
                        
        axes_label: boolean
            whether labels of axes are neeeded
        
        alpha: float, default 0.25
            indicates transparency
            
        savefig: str
            path for exported file
            
        """    
        if score is None:
            raise ValueError("!! Indicate score such as responase score !!")
        if len(focus)==0:
            focus = list(score.columns)
            if len(focus) > 10:
                if plot_all==False:
                    raise ValueError("!! Too many samples: turn on plot_all to plot all data !!")
        feature = list(score.index)
        abs_score = np.abs(score)
        if absolute:
            score = abs_score
        if overlay:
            if type(color)!=list:
                color = []
            dat = RadarChart()
            axes = []
            for i,f in enumerate(focus):
                if len(color)==0:
                    color = [""]*len(focus)
                else:
                    div,mod = divmod(len(focus),len(color))
                    color = color*div + color[:mod]
                temp = score[f].values
                temp2 = abs_score[f].sort_values(ascending=False)
                if n_labeled is None:
                    l2d = dat.plot(values=temp,features=feature,label=f,color=color[i],**kwargs)
                else:
                    shown = list(temp2.index)[:n_labeled]
                    feature2 = [v if v in shown else "" for v in feature]
                    l2d = dat.plot(values=temp,features=feature2,label=f,color=color[i],**kwargs)
                axes.append(l2d)
            dat.close(savefig=savefig,dpi=dpi,handles=axes,labels=focus)
        else:
            if type(color)!=str:
                raise TypeError("!! Color should be string when overlay is False !!")
            for f in focus:
                if figsize is None:                
                    figsize = (4,6) # hard coding to align the sizes of plots
                dat = RadarChart(figsize=figsize)
                temp = score[f].values
                temp2 = abs_score[f].sort_values(ascending=False)
                if n_labeled is None:
                    dat.plot(values=temp,features=feature,label=f,color=color,**kwargs)
                else:
                    shown = list(temp2.index)[:n_labeled]
                    feature2 = [v if v in shown else "" for v in feature]
                    dat.plot(values=temp,features=feature2,label=f,color=color,**kwargs)
                dat.close(savefig,dpi)


def _normal(X):
    return (X - np.mean(X,axis=0))/np.std(X,axis=0)


def _prep_label(n_total,n_focus):
    label = []
    ap = label.append
    for i in range(n_total):
        if i < n_focus:
            ap("P{}V".format(i + 1))
        else:
            ap("P{}".format(i + 1))
    return label





# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator o-- EstPlot

@author: tadahaya

"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# abstract class
class EstPlot():
    def __init__(self):
        pass

    def plot(self):
        raise NotImplementedError        

# concrete class
class ScreePlot(EstPlot):
    def plot(self,true_contr,rand_contr,fontsize=12,color="royalblue",markersize=8,
             linewidth=2,xlabel="factors",ylabel="eigen value",title="",
             focus=5,fileout="",dpi=100,figsize=(),**kwargs):
        """
        visualize a result of parallel analysis

        Parameters
        ----------
        res: dataframe
            a result file of enrichment analysis
            
        """
        x = np.arange(1,len(true_contr) + 1,1).astype(int)
        if len(figsize) > 0:
            fig = plt.figure(figsize=figsize)
        else:
            fig = plt.figure()
        ax = fig.add_subplot(111)
        if len(xlabel) > 0:
            ax.set_xlabel(xlabel,fontsize=fontsize)
        if len(ylabel) > 0:
            ax.set_ylabel(ylabel,fontsize=fontsize)
        if len(title) > 0:
            ax.set_title(title,fontsize=fontsize)
        ax.tick_params(labelsize=fontsize)
        if len(rand_contr)!=0:
            ax.plot(x,rand_contr,marker="o",markersize=markersize,
                    linewidth=linewidth,color="grey",markerfacecolor="grey",
                    markeredgecolor="grey",label="null")
        ax.plot(x,true_contr,marker="o",markersize=markersize,
                linewidth=linewidth,color=color,markerfacecolor=color,
                markeredgecolor=color,label="data")
        ax.legend(loc="best",fontsize=fontsize)
        if len(fileout) > 0:
            plt.savefig(fileout,bbox_inches="tight",dpi=dpi)
        plt.tight_layout()
        plt.show()


# concrete class
class AccPlot(EstPlot):
    def plot(self,contribution,nfactor=None,fontsize=12,color="royalblue",markersize=8,
             linewidth=2,xlabel="factors",ylabel="cumulative contribution",title="",
             focus=5,fileout="",dpi=100,figsize=(),**kwargs):
        """
        visualize cumulative contribution
            
        """
        x = np.arange(1,len(contribution) + 1,1).astype(int)
        total_contr = np.sum(contribution)
        contribution = contribution/total_contr
        acc = 0.0
        y = []
        for v in contribution:
            acc += v
            y.append(acc)
        if len(figsize) > 0:
            fig = plt.figure(figsize=figsize)
        else:
            fig = plt.figure()
        ax = fig.add_subplot(111)
        if len(xlabel) > 0:
            ax.set_xlabel(xlabel,fontsize=fontsize)
        if len(ylabel) > 0:
            ax.set_ylabel(ylabel,fontsize=fontsize)
        if len(title) > 0:
            ax.set_title(title,fontsize=fontsize)
        ax.tick_params(labelsize=fontsize)
        ax.plot(x,y,marker="o",markersize=markersize,
                linewidth=linewidth,color=color,markerfacecolor=color,
                markeredgecolor=color,label="contribution")
        if nfactor is not None:
            ax.vlines(nfactor,min(y),max(y),linestyle='dashed',linewidth=linewidth,
                      color="grey",label="nfactor: {}".format(nfactor))
        ax.legend(loc="best",fontsize=fontsize)
        if len(fileout) > 0:
            plt.savefig(fileout,bbox_inches="tight",dpi=dpi)
        plt.tight_layout()
        plt.show()



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- PreProcessor o-- PreProcess o-- DEG

@author: tadahaya
"""
import numpy as np
import pandas as pd

from .._utils import statistics as st
from .._utils import normalizer as nor


class DEGIqr():
    def __init__(self):
        pass

    def __vec2tpl(self,mtx,fold=2.0,method="iqr",nmin=None,nmax=None):
        """ convert dataframe into tuple of tags """
        sample_name = list(mtx.columns)
        n_sample = len(sample_name)
        n_feature = len(mtx.index)
        if nmin is None:
            nmin = 15
        if nmax is None:
            nmax = int(0.01*n_feature)
        if method=="std":
            upper,lower = st.outlier_std(mtx=mtx,fold=fold,axis=0)
        else:
            upper,lower = st.outlier_iqr(mtx=mtx,fold=fold,axis=0)
        res = []
        ap = res.append
        for i in range(n_sample):
            temp = mtx.iloc[:,i].sort_values(ascending=False)
            up_val = upper[i]
            low_val = lower[i]
            temp_l = list(temp.index)
            upper_tag = set(temp[temp > up_val].index)
            lower_tag = set(temp[temp < low_val].index)
            n_up = len(upper_tag)
            n_low = len(lower_tag)
            if n_up > nmax:
                upper_tag = set(temp_l[:nmax])
            elif n_up < nmin:
                upper_tag = set(temp_l[:nmin])
            if n_low > nmax:
                lower_tag = set(temp_l[-nmax:])
            elif n_low < nmin:
                lower_tag = set(temp_l[-nmin:])
            ap((upper_tag,lower_tag))
        return res


    def func(self,data,fold=3.0,method="iqr",nmin=None,nmax=None,
             raw=True,sep="_",position=0,control=""):
        """
        convert dataframe of response profiles into dict of tags

        Parameters
        ----------
        data: dataframe
            feature x sample matrix
            Groups of samples should be indicated like "treatment_1", "treatment_2", etc.

        fold: float
            determine threshold of outliers

        method: str
            "std" or "iqr"

        nmin,nmax: int
            indicate the minimum and maximum No. of genes in each tag
        
        raw: boolean
            indicates whether raw data is set or not
            if True, the data will be converted before tag making

        sep: str, default "_"
            separator for sample name
            
        position: int, default 0
            indicates position of sample name such as drug

        control: str
            indicates the control column name

        """
        if raw:
            check = [c for c in list(data.columns) if control not in c]
            check = [c for c in check if sep not in c]
            if len(check)==0: # check whether consensus signature can be applied
                data = nor.consensus_sig(data,sep=sep,position=position)
                data = nor.madz(data,control=control)
                data = nor.consensus_sig(data,sep=sep,position=position)
                col = [v.split(sep)[0] for v in list(data.columns)]
                data.columns = col
                data = data.T.groupby(level=0,axis=0).mean().T
            else:
                raise ValueError("!! Sample names should include a separator like '_' to indicate groups !!")
        temp = self.__vec2tpl(data,fold,method,nmin,nmax)
        dic = dict(zip(list(data.columns),temp))
        return dic


class DEGPval():
    def __init__(self):
        raise NotImplementedError

    def func(self,data):
        """
        get differentially expressed genes

        Parameters
        ----------
        data: array
            feature x data (ex. gene x sample) array

        """
        raise NotImplementedError




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

path handler

@author: tadahaya
"""
from pathlib import Path
import os

def get_base():
    """ get base path """
    return os.path.dirname(__file__)


def get_filepath(p,extension="txt"):
    """ obtain paths and file names of files in the indicated directory """    
    p2 = Path(p)
    paths = list(map(lambda x: x.as_posix(),list(p2.glob("*.{}".format(extension)))))
    names = [v.split("/")[-1].replace(".{}".format(extension),"") for v in paths]
    return paths,names





# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

module for plot

@author: tadahaya

"""

import pandas as pd
import numpy as np
from scipy.stats import rankdata
import matplotlib.pyplot as plt
import seaborn as sns
import csv
from pathlib import Path

class RadarChart():
    def __init__(self,figsize=None):
        """ prep polar chart """
        if figsize is not None:
            self.fig = plt.figure(figsize=figsize)
        else:
            self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111,polar=True)
        plt.gca().spines['polar'].set_color('lightgray')
        plt.gca().spines['start'].set_color('lightgray')
        plt.gca().spines['end'].set_color('lightgray')
        plt.gca().spines['inner'].set_color('lightgray')
        plt.grid(color='lightgray')
        self.feature = None # list
        self.rlim = None # tuple


    def plot(self,values,features,label="",limit=None,color="",
            alpha=0.25,markersize=5,axes_label=True,labelsize=10):
        """
        radar chart
        
        Parameters
        ----------
        values: array
            values to be plotted

        features: list
            list of features to be plotted
        
        label: str
            sample name
            
        limit: tuple
            determine min and max values of plot: (min,max)
        
        color: str
            determine the color of plot
                
        markersize: int
            determine size of markers
            
        labelsize: int
            determine size of axis-labels
            
        ticksize: int
            deterimne font size of polar axis
                        
        axes_label: boolean
            whether labels of axes are neeeded
        
        alpha: float, default 0.25
            indicates transparency
                        
        """    
        # prep
        if len(values)!=len(features):
            raise ValueError("!! len(values) and len(features) should be equal !!")
        angles = np.linspace(0,2 * np.pi, len(features) + 1,endpoint=True)
        maxvalue = np.max(values)
        minvalue = np.min(values)
        values = np.concatenate((values,[values[0]])) # since it rounds once

        # limit
        if limit==None:
            limit = (minvalue + minvalue/len(features),maxvalue + maxvalue/len(features))
        else:
            limit = limit
        if self.rlim is None:
            self.rlim = limit
        else:
            if self.rlim[0] > limit[0]:
                self.rlim = (limit[0],self.rlim[1])
            if self.rlim[1] < limit[1]:
                self.rlim = (self.rlim[0],limit[1])

        # feature
        if self.feature is None:
            self.feature = features
        else:
            self.feature = [v if len(v) > 0 else self.feature[i] for i,v in enumerate(features)]

        # visualize
        self.ax.plot(angles,np.zeros((len(angles),)),"-",color="black",linewidth=.5) # 0 line
        if len(color)==0:
            l2d, = self.ax.plot(angles,values,"o-",markersize=markersize) # determines the outer frame
            self.ax.fill(angles,values,alpha=alpha) # fill the frame
        else:
            l2d, = self.ax.plot(angles,values,"o-",color=color,label=label,markersize=markersize) # determines the outer frame
            self.ax.fill(angles,values,alpha=alpha,color=color) # fill the frame
        self.ax.set_theta_zero_location('N', offset=0) # determines the location of zero 'North'
        self.ax.set_theta_direction(-1) # determines the rotation: clockwise
        if axes_label:
            self.ax.set_thetagrids(angles[:-1] * 180 / np.pi, self.feature, fontsize=labelsize) #label of axis
        else:
            self.ax.set_thetagrids(angles[:-1] * 180 / np.pi, [""]*len(self.feature)) #label of axis
        self.ax.set_rlim(self.rlim)
        return l2d


    def close(self,savefig="",dpi=100,handles=[],labels=[]):
        """
        close the canvas

        Parameters
        ----------
        savefig: str
            path for exported file

        dpi: int
            indicates the dpi of the output

        """
        if (len(handles) > 0) and (len(labels) > 0):
            plt.legend(handles=handles,labels=labels,loc="best")
        else:
            plt.legend(loc="lower center",bbox_to_anchor=(0.5,-0.25))
        plt.tight_layout()
        if len(savefig) > 1:
            plt.savefig(fname=savefig,dpi=dpi,format="tif",bbox_inches="tight")
        plt.show()
        plt.close(self.fig)


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 23 13:05:34 2019

DePro o-- Estimator o-- Parallel

@author: tadahaya
"""
import pandas as pd
import numpy as np
from numpy.random import normal
from scipy.stats import bernoulli
import collections
from tqdm import trange
from sklearn.utils.extmath import randomized_svd


# concrete method
class Parallel():
    def __init__(self):
        self.nfactor = 0
        self.__true_contr = np.array([]) # for plot
        self.__rand_contr = np.array([]) # for plot
        self.__randomize = RandomFlip()


    def to_normal(self):
        """ change randomization strategy to that based on normal distribution """
        self.__randomize = RandomNormal()


    def to_shuffle(self):
        """ change randomization strategy to that based on shuffling data """
        self.__randomize = RandomShuffle()


    def to_flip(self):
        """ change randomization strategy to that based on flipping data """
        self.__randomize = RandomFlip()


    def get_contr(self):
        """
        get eigen values for contribution calculation and plot
        
        Return
        ------
        true_eigen_values: eigen values derived from the actual data
        random_eigen_values: eigen values derived from the representative random data
        
        """
        return self.__true_contr,self.__rand_contr


    def get_nfactor(self,X,perm:int=5,randomize:str="flip",**kwargs):
        """
        estimate the factor No. with parallel analysis
        randomization with Normal distribution

        Parameters
        ----------
        X: 2D array
            data x feature matrix

        perm: int
            indicate permutation No. for parallel analysis
        
        randomize: str
            indicate how to prepare randomized data for null contribution

        """
        ### preparation
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]        
        dat = PCAContr()
        dat.fit(X)
        self.__true_contr = dat.get_contr()

        ### contribution of randomized data
        if randomize=="flip":
            self.to_flip()
        elif randomize=="shuffle":
            self.to_shuffle()
        elif randomize=="normal":
            self.to_normal()
        else:
            raise KeyError("!! Wrong randomize: choose 'flip', 'shuffle', or 'normal' !!")
        rand_contr = self.__randomize.calc_random(X,perm)

        ### determine the threshold
        checker = self.__true_contr - rand_contr
        high = np.sum(checker > 0,axis=1)
        self.nfactor = _mode_higher(high)
        high_idx = [i for i,v in enumerate(high) if v==self.nfactor]
        self.__rand_contr = rand_contr[np.min(high_idx)]
        return self.nfactor


class RandomNormal():
    """ randomization based on normal distribution """
    def calc_random(self,X,perm:int):
        """ calculate the contribution of randomized data """
        X_c = X.copy()
        Xshape = X_c.shape
        myu = np.mean(np.mean(X_c,axis=0))
        scale = np.mean(np.std(X_c,ddof=1,axis=0))
        rand_contr = []
        ap = rand_contr.append
        for i in trange(perm):
            Xrand = normal(myu,scale,Xshape)
            
            ### get contribution
            dat = PCAContr()
            dat.fit(Xrand)
            temp_contr = dat.get_contr()
            ap(temp_contr)
        return np.array(rand_contr)


class RandomShuffle():
    """ randomization based on shuffling data """
    def calc_random(self,X,perm:int):
        """ calculate the contribution of randomized data """
        X_c = X.copy()
        rand_contr = []
        ap = rand_contr.append
        for i in trange(perm):
            rand_data = []
            ap2 = rand_data.append
            for v in X_c:
                temp = v.copy()
                np.random.shuffle(temp)
                ap2(temp)
            Xrand = np.array(rand_data)
            
            ### get contribution
            dat = PCAContr()
            dat.fit(Xrand)
            temp_contr = dat.get_contr()
            ap(temp_contr)
        return np.array(rand_contr)


class RandomFlip():
    """ randomization based on flipping data """
    def calc_random(self,X,perm:int):
        """ calculate the contribution of randomized data """
        X_c = X.copy()
        nfeat = X.shape[1] # assume that data size > No. of features
        ndata = X.shape[0]        
        rand_contr = []
        ap = rand_contr.append
        for i in trange(perm):
            be_rand = np.array([bernoulli.rvs(p=0.5,size=ndata) for j in range(nfeat)]).T*2 - 1
            Xrand = X_c*be_rand

            ### get contribution
            dat = PCAContr()
            dat.fit(Xrand)
            temp_contr = dat.get_contr()
            ap(temp_contr)
        return np.array(rand_contr)


class PCAContr():
    def __init__(self):
        self.X = np.array([[],[]])
        self.n = 0
        self.p = 0
        self.lmd = np.array([])
    
    def fit(self,X,n_components=None):
        """
        X: 2d array
            sample x feature matrix (n x p)
            should be normalized in row (feature)
        
        """
        self.n = X.shape[0]
        self.p = X.shape[1]
        if n_components is None:
            n_components = np.min((self.n,self.p))
        U,S,V = randomized_svd(X,np.min(X.shape),transpose=False)
        lmd = S**2/self.n # the effect of sqrt(n) is reflected on sigma
        self.lmd = lmd[:n_components]

    def get_contr(self):
        return self.lmd


def _mode_higher(array):
    """ return mode which is highest when multiple hits """
    res = collections.Counter(array).most_common()
    mod = res[0]
    res = [v[0] for v in res if v[1]==mod[1]]
    return np.max(res)